
--f = fopen "E:\\00713a4f.cmdl" "rb"
--f = fopen "E:\\00b7a02e.cmdl" "rb"
fsource = GetOpenFileName \
caption:"Select cmdl File" \
types: "cmdl(*.cmdl)|*.cmdl|All files (*.*)|*.*|"
if fsource != undefined then (
	
	f = fopen fsource "rb"
	clearlistener()

	fn readBElong fstream = (
		long = readlong fstream
		long = bit.swapBytes long 1 4
		long = bit.swapBytes long 2 3
		return long
	)

	fn readBEshort fstream = (
		short = readshort fstream --#unsigned
		short = bit.swapBytes short 1 2
		return short
	)

	fn readHalfFloat fstream = (
		hf=readBEshort fstream
		if (hf == 0) then return 0
		sign = bit.get hf 16
		exponent = (bit.shift (bit.and hf (bit.hexasint "7C00")) -10) as integer - 16
		fraction = bit.and hf (bit.hexasint "03FF")
		if sign==true then sign = 1 else sign = 0
		exponentF = exponent + 127
		outputAsFloat = bit.or (bit.or (bit.shift fraction 13) \
		(bit.shift exponentF 23)) (bit.shift sign 31)
		return bit.intasfloat outputasfloat*2
	)
		
	fn ReadBEfloat fstream = (
		fpt=readfloat fstream
		itger = bit.floatAsInt fpt
		hih = bit.intashex itger
		while hih.count < 8 do hih = "0" + hih
		shn = (substring hih 7 2) + \
		(substring hih 5 2) + \
		(substring hih 3 2) + \
		(substring hih 1 2)
		bit.intAsFloat (bit.hexasint shn)
	)

	fn ReadType fstream = (
		local str = ""
		for i = 1 to 4 do (
			str+= bit.IntAsChar(Readbyte fstream)
		)
		return str
	)

	types 		 =#()
	oBuffer 		 =#()
	uvBuffer		 =#()
	faceBuffer 	 =#()
	boneBuffer 	 =#()
	vertexBuffer =#()
	normalBuffer =#()
	weightBuffer =#()
	bIdxBuffer 	 = #()

	struct CmdlHeader (
		magic, --MODL
		version , --1
		offset,
		numType
	)

	struct CmdlType (
		str,
		id,
		flag,
		offset,
		size
	)

	struct CmdlBone (
		posX,
		posY,
		posZ,
		rotX,
		rotY,
		rotZ,
		parent,
		padding,
		weightOffset,
		weightLength,
		unkOffset,
		unkLength,
		numUnk,
		unk=#(),
		weight1,
		weight2,
		weight3,
		weight4,
		weight5,
		weight6,
		weight7,
		weight8,
		padding2,
		boneGroup,
		boneIdx
	)
	--read header--
	header = CmdlHeader()

	header.magic		= readType f
	header.version		= readBElong f
	header.offset		= readBElong f 
	header.numType	= readBElong f 


	--read type table--
	for i = 1 to header.numType do (
		type = CmdlType()
		
		type.str		= readType f
		type.id		= readBEshort f
		type.flag	= readBEshort f
		type.offset	= readBElong f
		type.size 	= readBElong f
		
		append types type
	)	
	
	--read types--
	for i = 1 to types.count do (
		fseek f (types[i].offset + 0x0C) #seek_set
		
		if (types[i].str == "POS0") then (
			for x=1 to types[i].size/16 do (
				vx		= readBEfloat f 
				vy		= readBEfloat f 
				vz		= readBEfloat f
				vw	= readBEfloat f
				
				append vertexBuffer[vx,vy,vz]
			)
		)
		
		else if (types[i].str == "NRM0") then (
			for x=1 to types[i].size/4 do (
				bitstream = readBElong f
		
				nx = bit.and (bit.shift bitstream 0) 0x07FF
				nx = bit.or	nx (0 - (bit.and nx 0x0400))
				nx = nx / 1023 as float
					
				ny = bit.and (bit.shift bitstream -11) 0x07FF
				ny = bit.or	ny (0 - (bit.and ny 0x0400))	
				ny = ny / 1023 as float
					
				nz = bit.and (bit.shift bitstream -22) 0x03FF
				nz = bit.or	nz (0 - (bit.and nz 0x0200))		
				nz = nz / 511 as float
					
				append normalBuffer [-nx, -ny, -nz]
			)
		)
		
		
		else if (types[i].str == "TEX0") then (
			for x=1 to types[i].size/4 do (
				tu = ReadHalfFloat f 
				tv = ReadHalfFloat f
				append uvBuffer[tu,tv, 0]
			)
		)
		
		else if (types[i].str == "OIDX") then (
			for x=1 to types[i].size/12 do (
				fa = readBElong f + 1
				fb = readBElong f + 1 
				fc = readBElong f + 1
				
				append oBuffer[fa, fb, fc]
			)
		)
		
		else if (types[i].str == "BONI") then (
			for x=1 to types[i].size/4 do (
				fa = readbyte f
				fb = readbyte f
				fc = readbyte f
				fd = readbyte f
				append bIdxBuffer [fa, fb, fc, fd]
			)
		)
		
		else if (types[i].str == "BONW") then (
			for x=1 to types[i].size/8 do (
				w1 = ReadHalfFloat f
				w2 = ReadHalfFloat f 
				w3 = ReadHalfFloat f 
				w4 = ReadHalfFloat f 
				append weightBuffer[w1, w2, w3, w4]
			)
		)
		
		else (
			print "unsupported type"
		)
		
	)
	
	--read MODL--
	fseek f (header.offset + 0x0C) #seek_set
	faceLength =  readBElong f
	
	for x=1 to faceLength/3 do (
		fa = (readBElong f ) +1
		fb = (readBElong f ) +1
		fc = (readBElong f ) +1
		
		append faceBuffer[fa,fb,fc]
	)
	
	fseek f 0x04 #seek_cur
	
	boneInfoLength = readBElong f
	
	for i=1 to boneInfoLength do (
		boneIdx = CmdlBone()
		
		boneIdx.posX 				= readBEfloat f
		boneIdx.posY 				= readBEfloat f
		boneIdx.posZ 				= readBEfloat f
		boneIdx.rotX 				= readBEfloat f
		boneIdx.rotY 				= readBEfloat f
		boneIdx.rotZ 				= readBEfloat f
		boneIdx.parent			= readBEshort f + 1
		boneIdx.padding 			= readbyte f
		boneIdx.weightOffset 	= readBElong f
		boneIdx.weightLength	= readBElong f
		boneIdx.unkOffset		= readBElong f
		boneIdx.unkLength		= readBElong f
		boneIdx.numUnk 			= readBElong f
		
		for x = 1 to boneIdx.numUnk do (
			unknown = readBElong f
			append boneIdx.unk unknown
		)
		
		boneIdx.weight1			= readbyte f
		boneIdx.weight2			= readbyte f
		boneIdx.weight3			= readbyte f
		boneIdx.weight4			= readbyte f
		boneIdx.weight5			= readbyte f
		boneIdx.weight6			= readbyte f
		boneIdx.weight7			= readbyte f
		boneIdx.weight8			= readbyte f
		boneIdx.padding2			= readBElong f
		boneIdx.boneGroup		= readBElong f
		boneIdx.boneIdx			= readBElong f
		
		append boneBuffer boneIdx
	)
	
		for i = 1 to boneBuffer.count do (
			
			--if getNodeByName (boneBuffer[i].boneGroup as string) != undefined then continue
			
			px =	boneBuffer[i].posX;	 py = boneBuffer[i].posY;		pz = boneBuffer[i].posZ;
			rx =	boneBuffer[i].rotX; 	 ry = boneBuffer[i].rotY;		rz = boneBuffer[i].rotZ;
			
			BonePos = [(px + rx) / 2 ,(py + ry) / 2 ,(pz + rz) / 2]
			--BonePos = [px, py, px]
			bne = dummy showLinks:true showLinksOnly:true
			bne.name = append (boneBuffer[i].boneGroup as string) (append "-" (boneBuffer[i].boneIdx as string))
			--bne.name = boneBuffer[i].boneGroup as string
			--bne.scale *= 16
			
			--parent = boneBuffer[i].boneGroup - 1
			--if (parenta - 1  != -1) then bne.parent = getNodeByName (parenta as string)
			in coordsys parent bne.pos = BonePos
			--scaleToBoundingBox bne boneBuffer[i]
		
	)
	
	print boneBuffer
	print (append "boneBuffer: " (bonebuffer.count as string))
	print (append "weightbuffer: "  (weightBuffer.count as string))
	print (append "bidxbuffer: "  (bIdxBuffer.count as string))
	print (append "uvbuffer: "  (uvBuffer.count as string))
	print (append "vertexbuffer: "  (vertexBuffer.count as string))
	print (append "obuffer: "  (obuffer.count as string))
	print "bIdxBuffer"
	print bIdxBuffer
	print "weightBuffer"
	print weightBuffer
	print types
	print(bit.intAsHex(ftell f))

	msh = mesh vertices:vertexBuffer faces:faceBuffer
    msh.numTVerts = uvBuffer.count
	buildTVFaces msh

	for j = 1 to uvBuffer.count do setTVert msh j uvBuffer[j]		
	for j = 1 to faceBuffer.count do setTVFace msh j faceBuffer[j]
	for j = 1 to normalBuffer.count do setNormal msh j normalBuffer[j]

	if weightBuffer.count != 0 then (
		max modify mode
		skinMod = skin ()
		addModifier msh skinMod
		select msh
		
		for i =1 to boneBuffer.count do (
			boneGroup = boneBuffer[i].boneGroup
			boneID = boneBuffer[i].boneIdx
			boneStr = append (boneGroup as string) (append "-" (boneID as string))
			maxBone = getnodeByName (boneStr)
			skinOps.addBone skinMod maxBone 0
			
			modPanel.setCurrentObject skinMod
			boneNames = for i=1 to (skinOps.GetNumberBones skinMod) collect (skinOps.GetBoneName skinMod i 0)
			boneIndex = findItem boneNames boneStr	
				
			for j = boneBuffer[i].weightOffset + 1  to (boneBuffer[i].weightOffset + boneBuffer[i].weightLength) do (
				k = bIdxBuffer[j][1] + 1
				skinOps.SetVertexWeights skinMod j  boneIndex (weightBuffer[j][k])
			)			
		)
	)
	
	gc()
	fclose f
)
else (
   clearlistener()
)